package POWER::Epoll;

use version; $VERSION = qv('2.0.2');    # update POD & Changes & README

use warnings;
use strict;
use Carp;

my @subs    = qw(
    epoll_create
    epoll_add
    epoll_mod
    epoll_del
    epoll_wait
);
my @consts  = qw(
    EPOLLIN
    EPOLLOUT
    EPOLLPRI
    EPOLLERR
    EPOLLHUP
    EPOLLET
    EPOLLONESHOT
);

sub import {
    my $pkg = caller(0);
    no strict 'refs';
    *{$pkg.'::'.$_} = \&{'_'.$_} for @subs, @consts;
}

# XXX  C-   _     
#   epoll.h .
use Inline C => <<'END_OF_C';
#include <sys/epoll.h>
#include <sys/syscall.h>
#include <math.h>

// workaround to work on system with linux-headers-2.4 but running kernel-2.6
#define __NR_epoll_create	254
#define __NR_epoll_ctl      255
#define __NR_epoll_wait		256
#define SYS_epoll_create	__NR_epoll_create
#define SYS_epoll_ctl		__NR_epoll_ctl
#define SYS_epoll_wait		__NR_epoll_wait

#define MAXEVENTS 1000
static struct epoll_event events[MAXEVENTS];

int _epoll_create(int size) {
    int res;
	if ((res = syscall(SYS_epoll_create, size)) == -1)
        croak("epoll_create: %s\n", strerror(errno));
    return res;
}

void _epoll_add(int epfd, int fd, int events) {
	struct epoll_event ev;
	ev.data.fd = fd;
	ev.events  = events;
	if (syscall(SYS_epoll_ctl, epfd, EPOLL_CTL_ADD, fd, &ev) == -1)
        croak("epoll_add: %s\n", strerror(errno));
}

void _epoll_mod(int epfd, int fd, int events) {
	struct epoll_event ev;
	ev.data.fd = fd;
	ev.events  = events;
	if (syscall(SYS_epoll_ctl, epfd, EPOLL_CTL_MOD, fd, &ev) == -1)
        croak("epoll_mod: %s\n", strerror(errno));
}

void _epoll_del(int epfd, int fd) {
	struct epoll_event ev;
	ev.data.fd = fd;
	ev.events  = 0;
	if (syscall(SYS_epoll_ctl, epfd, EPOLL_CTL_DEL, fd, &ev) == -1)
        croak("epoll_del: %s\n", strerror(errno));
}

void _epoll_wait(int epfd, SV* timeout) {
	int n, nfds, timeout_ms;
	AV* results;
    Inline_Stack_Vars;
    Inline_Stack_Reset;
	if (SvOK(timeout)) {
        timeout_ms = (int)(SvNV(timeout) * 1000);
        if (timeout_ms < 0)
            croak("epoll_wait: timeout can't be negative\n");
    }
    else {
        timeout_ms = -1;
    }
	results = (AV *)sv_2mortal((SV *)newAV());
	nfds = syscall(SYS_epoll_wait, epfd, &events, MAXEVENTS, timeout_ms);
    if (nfds == -1 && errno != EINTR)
        croak("epoll_wait: %s\n", strerror(errno));
    Inline_Stack_Push(sv_2mortal(newSViv(nfds)));
	av_extend(results, nfds); // optimization
	for(n = 0; n < nfds; ++n) {
	    AV* av;
	    av = (AV *)sv_2mortal((SV *)newAV());
	    av_push(av, newSViv(events[n].data.fd));
	    av_push(av, newSViv(events[n].events));
	    av_push(results, newRV((SV *)av));
	}
    Inline_Stack_Push(sv_2mortal(newRV((SV *)results)));
    Inline_Stack_Done;
}

int _EPOLLIN() {
    return EPOLLIN;
}

int _EPOLLOUT() {
    return EPOLLOUT;
}

int _EPOLLPRI() {
    return EPOLLPRI;
}

int _EPOLLERR(){
    return EPOLLERR;
}

int _EPOLLHUP() {
    return EPOLLHUP;
}

int _EPOLLET() {
    return EPOLLET;
}

int _EPOLLONESHOT() {
    return EPOLLONESHOT;
}
END_OF_C


1; # Magic true value required at end of module
__END__

=head1 NAME

POWER::Epoll - Perl extension for epoll(4)


=head1 VERSION

This document describes POWER::Epoll version 2.0.2


=head1 SYNOPSIS

 use POWER::Epoll;

 my $epfd = epoll_create(100);

 epoll_add($epfd, fileno(SOMEFD1), EPOLLET|EPOLLIN);
 epoll_add($epfd, fileno(SOMEFD2), EPOLLET|EPOLLIN|EPOLLOUT);
 epoll_mod($epfd, fileno(SOMEFD2), EPOLLET|EPOLLOUT);
 epoll_del($epfd, fileno(SOMEFD2));

 ($status, $events) = epoll_wait($epfd, $timeout);
 if ($status == -1) {
    # epoll_wait was interrupted by signal (EINTR)
 }
 elsif ($status == 0) {
    # epoll_wait was exited because no events happens until $timeout
 }
 else {
    # process events in @$events
 }


=head1 DESCRIPTION

Interface for epoll(4) in linux kernels 2.6+.

Use syscall(), so glibc compiled with linux headers 2.6+ isn't required.


=head1 EXPORT

Everything below is mandatory exported.

=over

=item Constants (see epoll_ctl(2) for descriptions) :

EPOLLIN EPOLLOUT EPOLLPRI EPOLLERR EPOLLHUP EPOLLET EPOLLONESHOT


=item Functions:

epoll_create epoll_add epoll_mod epoll_del epoll_wait


=back

=head1 INTERFACE 

=over

=item B<epoll_create>( $size )

See L<epoll_create(2)>.

Return: $epfd.


=item B<epoll_add>( $epfd, $fileno, $events )

See L<epoll_ctl(2)> with op=EPOLL_CTL_ADD.

Return: nothing.


=item B<epoll_mod>( $epfd, $fileno, $events )

See L<epoll_ctl(2)> with op=EPOLL_CTL_MOD.

Return: nothing.


=item B<epoll_del>( $epfd, $fileno )

See L<epoll_ctl(2)> with op=EPOLL_CTL_DEL.

Return: nothing.


=item B<epoll_wait>( $epfd, $timeout )

See L<epoll_wait(2)>. $timeout should be in seconds, not in milliseconds,
but can be fractional. Set $timeout to undef for unlimited timeout.
Set $timeout to 0 to return immediatelly with all events which already happens.

Return list: ($wait_status, [[$fileno1, $events1], [$fileno2, $events2], ...]).
$wait_status meaning:

 -1     EINTR happens (any other error will throw exception)
 0      no events happens before $timeout
 >0     number of file descriptors with some events happens

=back

=head1 DIAGNOSTICS

=over

=item C<< epoll_create: $! >>

See L<epoll_create(2)>.


=item C<< epoll_add: $! >>

=item C<< epoll_mod: $! >>

=item C<< epoll_del: $! >>

See L<epoll_ctl(2)> with corresponding I<op> value.


=item C<< epoll_wait: timeout can't be negative >>

Second epoll_wait() param must be one of: undef, 0 or positive fractional
number.


=item C<< epoll_wait: $! >>

See L<epoll_wait(2)>. This exception will be throws for any error except
EINTR.


=back

=head1 CONFIGURATION AND ENVIRONMENT

POWER::Epoll requires no configuration files or environment variables.


=head1 DEPENDENCIES

 version
 Inline


=head1 INCOMPATIBILITIES

None reported.


=head1 BUGS AND LIMITATIONS

No bugs have been reported.

Please report any bugs or feature requests to
C<bug-power-epoll@rt.cpan.org>, or through the web interface at
L<http://rt.cpan.org>.


=head1 AUTHOR

 Alex Efros  <powerman@asdfgroup.com>
 Nikita Savin <nikita@asdfGroup.com>
 Ilya Persky <ilya_persky@yahoo.com>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2005, Alex Efros <powerman@asdfgroup.com>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
